# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

import pandas as pd
from time import sleep
from datetime import date, datetime

from tw_stock_plugin.object.stock_shareholdings import StockShareholdingsObject
from tw_stock_plugin.utils.response_handler import ResponseHandler
from tw_stock_plugin.constant import Domain


class StockShareholdings:
    def __init__(self):
        pass

    @staticmethod
    def _valid_date(date_, date_list):
        """
        check date format and if date in valid date list
        """
        if not isinstance(date_, date):
            raise TypeError('input date must be type of datetime.date')
        elif date_ not in date_list:
            raise ValueError('input date no available')
        else:
            return date_.strftime('%Y%m%d')

    @staticmethod
    def _get_available_date_list():
        url = f'{Domain.TDCC}/smWeb/QryStockAjax.do'
        headers = {'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) '
                                 'AppleWebKit/537.36 (KHTML, liek Gecko) Chrome/34.0.1847.131 Safari/537.36'}
        payload = {
            'REQ_OPR': 'qrySelScaDates'
        }
        while True:
            response = ResponseHandler.post(url=url, payload=payload, headers=headers)
            date_list = response.json()
            date_list = [datetime.strptime(date_, '%Y%m%d').date() for date_ in date_list]
            if date_list:
                return date_list
            print('fetching date list error, retry after 3 seconds...')
            sleep(3)

    @staticmethod
    def _fetch_single_data(code, date_):
        url = f'{Domain.TDCC}/smWeb/QryStockAjax.do'
        headers = {'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) '
                                 'AppleWebKit/537.36 (KHTML, liek Gecko) Chrome/34.0.1847.131 Safari/537.36'}
        payload = {
            'scaDates': date_,
            'scaDate': date_,
            'SqlMethod': 'StockNo',
            'StockNo': code,
            'radioStockNo': code,
            'StockName': '',
            'REQ_OPR': 'SELECT',
            'clkStockNo': code,
            'clkStockName': '',
        }
        while True:
            response = ResponseHandler.post(url=url, payload=payload, headers=headers)
            if '無此資料' not in response:
                return response
            print('fetching data error, retry after 3 seconds...')
            sleep(3)

    def get_by_query(self, code, date_):
        """
        fetch shareholdings data with specific code and date
        :param code:
        :param date_:
        :return:
        """
        date_list = self._get_available_date_list()
        date_ = self._valid_date(date_=date_, date_list=date_list)

        shareholdings_dict = dict()
        response = self._fetch_single_data(code=code, date_=date_)
        df = pd.read_html(response.text)[0]
        df = df.dropna()
        columns = ['date', 'code', 'index', 'number_of_shares', 'number_of_shareholders', 'total_shares',
                   'percentage_over_total_shares']
        df = df.iloc[1:]
        for data in df.values.tolist():
            data = [date_, code] + data
            stock_shareholdings = StockShareholdingsObject(**(dict(zip(columns, data))))
            if not stock_shareholdings.index:
                continue
            shareholdings_dict[stock_shareholdings.index] = stock_shareholdings
        return shareholdings_dict

    @staticmethod
    def get_newest():
        """
        fetch newest shareholdings data from official website
        :return:
        """
        shareholdings_dict = dict()
        url = f'{Domain.SMART_TDCC}/opendata/getOD.ashx'
        headers = {'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) '
                                 'AppleWebKit/537.36 (KHTML, liek Gecko) Chrome/34.0.1847.131 Safari/537.36'}
        params = {
            'id': '1-5'
        }
        response = ResponseHandler.get(url=url, params=params, headers=headers)
        df = pd.read_csv(response.url)
        columns = ['date', 'code', 'index', 'number_of_shareholders', 'total_shares',
                   'percentage_over_total_shares']
        for data in df.values.tolist():
            stock_shareholdings = StockShareholdingsObject(**(dict(zip(columns, data))))
            if not stock_shareholdings.number_of_shares:
                continue
            if stock_shareholdings.code not in shareholdings_dict:
                shareholdings_dict[stock_shareholdings.code] = dict()
            shareholdings_dict[stock_shareholdings.code].update({stock_shareholdings.index: stock_shareholdings})
        return shareholdings_dict



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
import os
import glob
import pandas as pd

from tw_stock_plugin import UpdateStock
from tw_stock_plugin.config import Config
from tw_stock_plugin.object.stock_info import StockInfoObject


class StockInfo:
    """
    TODO
        https://dts.twse.com.tw/opendata/t187ap03_L.csv
        https://dts.twse.com.tw/opendata/t187ap03_O.csv
    for getting basic stock info
    """

    def __init__(self):
        self._csv_dir = Config.STOCK_CSV_DIR
        self._stock_data = self._load_stocks()

    def _get_stock_csv_files(self):
        """
        check if csv files exists, if so, return csv files, else, crawling data from website
        :return:
        """
        if not (os.path.isdir(self._csv_dir) and len(os.listdir(self._csv_dir)) == 3):
            UpdateStock.main()
        return glob.glob(f'{self._csv_dir}/*.csv')

    def _load_stocks(self):
        """
        init stock data when object created
        :return:
        """
        stocks = dict()
        csv_files = self._get_stock_csv_files()
        for csv in csv_files:
            df = pd.read_csv(csv, index_col=0)
            df['證券代號'] = df['證券代號'].astype(str)
            df = df.fillna('')
            for data in df.values.tolist():
                stock_info = StockInfoObject(*data)
                stocks[stock_info.code] = stock_info
        return stocks

    def get(self, code=None):
        """
        get stock data by stock code
        :param code:
        :return:
        """
        if code:
            if code not in self._stock_data:
                raise ValueError('stock code not exists')
            return self._stock_data[code]
        else:
            return self._stock_data



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from tw_stock_plugin.utils.schema import SchemaPattern


class StockInfoObject:
    def __init__(self, code, name, ISIN_code, listed_date, category, industry, CFI_code, type):
        """
        :param code: 股票代碼
        :param name: 股票名稱
        :param ISIN_code: 國際證券辨識號碼(ISIN Code)
        :param listed_date: 上市/上櫃/興櫃 日
        :param category: 市場別
        :param industry: 產業別
        :param CFI_code: CFICode
        :param type: 證券類別
        """
        self.code = code
        self.name = name
        self.ISIN_code = ISIN_code
        self.listed_date = listed_date
        self.category = category
        self.industry = industry
        self.CFI_code = CFI_code
        self.type = type
        self._format_value()
        # self._valid_schema()

    def _format_value(self):
        for key, value in self.__dict__.items():
            if not value:
                setattr(self, key, None)

    def _valid_schema(self):
        SchemaPattern.StockInfoSchema.validate(self.__dict__)



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

import os
import requests
import pandas as pd
from time import sleep

from tw_stock_plugin.config import Config
from tw_stock_plugin.constant import Domain
from tw_stock_plugin.regex_pattern import StockPattern
from tw_stock_plugin.utils.response_handler import ResponseHandler


class UpdateStock:
    _STR_MODE_DICT = {
        'stock_exchange': 2,
        'over_the_counter': 4,
        'emerging_stock': 5
    }

    @staticmethod
    def _get_response(url):
        """
        :param url:
        :return:
        """
        response = requests.get(url)
        return response

    @classmethod
    def main(cls):
        """
        :return:
        """
        # create csv dir if dir not exists
        if not os.path.isdir(Config.STOCK_CSV_DIR):
            os.mkdir(Config.STOCK_CSV_DIR)

        print('UPDATING STOCK INFO')
        for category, str_mode in cls._STR_MODE_DICT.items():
            params = {
                'strMode': str_mode
            }
            response = ResponseHandler.get(url=f'{Domain.TAIWAN_STOCK_EXCHANGE_CORPORATION_ISIN}/isin/C_public.jsp',
                                           params=params)

            df = pd.read_html(response.text)[0]
            df_columns = df.iloc[0].values.tolist() + ['類別']
            new_data = list()
            tmp_type = None
            for data in df.values.tolist()[1:]:
                if not StockPattern.CODE_NAME_PATTERN.search(data[0]) and len(set(data)) == 1:
                    tmp_type = data[0]
                elif StockPattern.CODE_NAME_PATTERN.search(data[0]):
                    data.append(tmp_type)
                    new_data.append(data)
            df = pd.DataFrame(new_data, columns=df_columns)
            df['有價證券代號及名稱'] = df['有價證券代號及名稱'].apply(lambda x: x.replace('\u3000', ' '))
            code_name_df = pd.DataFrame(df['有價證券代號及名稱'].str.split(' ', 1).tolist(),
                                        columns=['證券代號', '證券名稱'])
            df = df.drop(['有價證券代號及名稱', '備註'], axis=1)
            df = pd.concat([code_name_df, df], axis=1)
            csv_path = os.path.join(Config.STOCK_CSV_DIR, f'{category}.csv')
            df.to_csv(csv_path)
            # to avoiding banned by twse website
            sleep(3)
        print('FINISHED')


if __name__ == '__main__':
    UpdateStock.main()



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from datetime import datetime
from tw_stock_plugin.core.stock_peratio import StockPERatio

# UpdateStock.main()

date_ = datetime(2020, 10, 30).date()

stock_p_e_ratio = StockPERatio(date_=date_)


print(stock_p_e_ratio.get_history(code='9962')[date_])



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from datetime import date

from tw_stock_plugin.core.stock_tools import StockTools
from tw_stock_plugin.object.stock_institutional_investors import InstitutionalInvestorsObject
from tw_stock_plugin.constant import Domain
from tw_stock_plugin.utils.response_handler import ResponseHandler


class StockInstitutionalInvestors:
    def __init__(self, date_):
        """
        :param date_: 查詢日期
        """
        self.date_ = date_
        self._valid_date()

    def _valid_date(self):
        """
        valid date format
        """
        if not isinstance(self.date_, date):
            raise TypeError('input date must be type of datetime.date')

    def _fetch_twse_data_all(self):
        """
        fetch all institutional investors data in specific date from twse website
        :return:
        """
        institutional_investors_dict = dict()
        url = f'{Domain.TAIWAN_STOCK_EXCHANGE_CORPORATION}/fund/T86'
        query_date = self.date_.strftime('%Y%m%d')
        params = {
            'response': 'json',
            'date': query_date,
            'selectType': 'ALL'
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        stats = json_data['stat']
        if not stats == 'OK':
            return None
        data_list = json_data['data']
        columns = ['code', 'name', 'foreign_mainland_area_buy', 'foreign_mainland_area_sell',
                   'foreign_mainland_area_diff', 'foreign_buy', 'foreign_sell', 'foreign_diff', 'trust_buy',
                   'trust_sell', 'trust_diff', 'proprietary_dealers_buy', 'proprietary_dealers_sell',
                   'proprietary_dealers_diff', 'hedge_dealers_buy', 'hedge_dealers_sell', 'hedge_dealers_diff',
                   'total_diff']
        removed_indices = {11}
        for data in data_list:
            data = [value for index, value in enumerate(data) if index not in removed_indices]
            stock_institutional_investors = InstitutionalInvestorsObject(**dict(zip(columns, data)))
            institutional_investors_dict[stock_institutional_investors.code] = stock_institutional_investors
        return institutional_investors_dict

    def _fetch_tpex_data_all(self):
        """
        fetch all institutional investors data in specific date from tpex website
        :return:
        """
        institutional_investors_dict = dict()
        url = f'{Domain.TAIPEI_EXCHANGE}/web/stock/3insti/daily_trade/3itrade_hedge_result.php'
        query_date = StockTools.ad_to_republic_era(date_=self.date_).replace('-', '/'),
        params = {
            'l': 'zh-tw',
            'd': query_date,
            'se': 'EW',
            't': 'D'
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        data_list = json_data['aaData']
        if not data_list:
            return None
        columns = ['code', 'name', 'foreign_mainland_area_buy', 'foreign_mainland_area_sell',
                   'foreign_mainland_area_diff', 'foreign_buy', 'foreign_sell', 'foreign_diff', 'trust_buy',
                   'trust_sell', 'trust_diff', 'proprietary_dealers_buy', 'proprietary_dealers_sell',
                   'proprietary_dealers_diff', 'hedge_dealers_buy', 'hedge_dealers_sell', 'hedge_dealers_diff',
                   'total_diff']
        removed_indices = {8, 9, 10, 20, 21, 22, 24}
        for data in data_list:
            data = [value for index, value in enumerate(data) if index not in removed_indices]
            stock_institutional_investors = InstitutionalInvestorsObject(**dict(zip(columns, data)))
            institutional_investors_dict[stock_institutional_investors.code] = stock_institutional_investors
        return institutional_investors_dict

    def get_all(self):
        """
        return all institutional investors data
        :return:
        """
        institutional_investors_dict = dict()
        twse_data = self._fetch_twse_data_all()
        tpex_data = self._fetch_tpex_data_all()
        if twse_data:
            institutional_investors_dict.update(twse_data)
        if tpex_data:
            institutional_investors_dict.update(tpex_data)
        return institutional_investors_dict



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

from datetime import date, datetime

from tw_stock_plugin.core.stock_tools import StockTools
from tw_stock_plugin.constant import Domain
from tw_stock_plugin.object.stock_peratio import StockPERatioObject
from tw_stock_plugin.utils.response_handler import ResponseHandler
from tw_stock_plugin.regex_pattern import PERatioPattern


class StockPERatio:
    def __init__(self, date_):
        """
        :param date_: 查詢日期
        """
        self.date_ = date_
        self._valid_date()

    def _valid_date(self):
        """
        valid date format
        """
        if not isinstance(self.date_, date):
            raise TypeError('input date must be type of datetime.date')

    @staticmethod
    def _translate_date(date_):
        """
        translate date from chinese to correct format
        :param date_:
        :return:
        """
        year, month, day = PERatioPattern.CHINESE_DATE_PATTERN.search(date_).groups()
        return f'{year}/{month}/{day}'

    def _fetch_twse_data_all(self):
        """
        fetch all p/e ratio, dividend yield and p/b ratio data in specific date from twse website
        :return:
        """
        p_e_ratio_dict = dict()
        url = f'{Domain.TAIWAN_STOCK_EXCHANGE_CORPORATION}/exchangeReport/BWIBBU_d'
        query_date = self.date_.strftime('%Y%m%d')
        params = {
            'response': 'json',
            'date': query_date,
            'selectType': 'ALL'
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        stats = json_data['stat']
        if not stats == 'OK':
            return None
        data_list = json_data['data']
        columns = ['code', 'name', 'yield_ratio', 'dividend_year', 'per', 'pbr', 'fiscal_year_quarter']
        for data in data_list:
            p_e_ratio = StockPERatioObject(**dict(zip(columns, data)))
            p_e_ratio_dict[p_e_ratio.code] = p_e_ratio
        return p_e_ratio_dict

    def _fetch_tpex_data_all(self):
        """
        fetch all p/e ratio, dividend yield and p/b ratio data in specific date from tpex website
        :return:
        """
        p_e_ratio_dict = dict()
        url = f'{Domain.TAIPEI_EXCHANGE}/web/stock/aftertrading/peratio_analysis/pera_result.php'
        query_date = StockTools.ad_to_republic_era(date_=self.date_).replace('-', '/'),
        params = {
            'l': 'zh-tw',
            'o': 'json',
            'd': query_date
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        data_list = json_data['aaData']
        if not data_list:
            return None
        columns = ['code', 'name', 'per', 'dividend_per_share', 'dividend_year', 'yield_ratio', 'pbr']
        for data in data_list:
            p_e_ratio = StockPERatioObject(**dict(zip(columns, data)))
            p_e_ratio_dict[p_e_ratio.code] = p_e_ratio
        return p_e_ratio_dict

    def get_all(self):
        """
        return all p/e ratio, dividend yield and p/b ratio data
        :return:
        """
        p_e_ratio_dict = dict()
        twse_data = self._fetch_twse_data_all()
        tpex_data = self._fetch_tpex_data_all()
        if twse_data:
            p_e_ratio_dict.update(twse_data)
        if tpex_data:
            p_e_ratio_dict.update(tpex_data)
        return p_e_ratio_dict

    def _fetch_twse_data_history(self, code):
        """
        get specific stock monthly p/e ratio, dividend yield and p/b ratio data from twse website
        :param code:
        :return:
        """
        p_e_ratio_dict = dict()
        url = f'{Domain.TAIWAN_STOCK_EXCHANGE_CORPORATION}/exchangeReport/BWIBBU'
        query_date = self.date_.strftime('%Y%m%d')
        params = {
            'response': 'json',
            'date': query_date,
            'stockNo': code
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        stats = json_data['stat']
        if not stats == 'OK':
            return None
        data_list = json_data['data']
        columns = ['yield_ratio', 'dividend_year', 'per', 'pbr', 'fiscal_year_quarter']

        for data in data_list:
            date_ = self._translate_date(data[0])
            date_ = StockTools.republic_era_to_ad(date_)
            date_ = datetime.strptime(date_, '%Y/%m/%d').date()
            p_e_ratio = StockPERatioObject(**(dict(zip(columns, data[1:]))))
            for attr in p_e_ratio.__dict__.copy():
                if attr not in columns:
                    delattr(p_e_ratio, attr)
            p_e_ratio_dict[date_] = p_e_ratio
        return p_e_ratio_dict

    def _fetch_tpex_data_history(self, code):
        """
        get specific stock monthly p/e ratio, dividend yield and p/b ratio data from tpex website
        :param code:
        :return:
        """
        p_e_ratio_dict = dict()
        url = f'{Domain.TAIPEI_EXCHANGE}/web/stock/aftertrading/peratio_stk/pera_result.php?'
        query_date = StockTools.ad_to_republic_era(date_=self.date_).replace('-', '/'),
        params = {
            'l': 'zh-tw',
            'd': query_date,
            'stkno': code
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        data_list = json_data['aaData']
        if not data_list:
            return None
        columns = ['per', 'yield_ratio', 'dividend_year', 'pbr']
        for data in data_list:
            date_ = StockTools.republic_era_to_ad(data[0])
            date_ = datetime.strptime(date_, '%Y/%m/%d').date()
            p_e_ratio = StockPERatioObject(**(dict(zip(columns, data[1:]))))
            for attr in p_e_ratio.__dict__.copy():
                if attr not in columns:
                    delattr(p_e_ratio, attr)
            p_e_ratio_dict[date_] = p_e_ratio
        return p_e_ratio_dict

    def get_history(self, code):
        """
        return monthly p/e ratio, dividend yield and p/b ratio data with specific stock code
        :param code:
        :return:
        """
        twse_data = self._fetch_twse_data_history(code=code)
        if twse_data:
            return twse_data

        tpex_data = self._fetch_tpex_data_history(code=code)
        if tpex_data:
            return tpex_data

        raise ValueError(f'{code} {self.date_} result not found')



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
import re


class DatePattern:
    REPUBLIC_ERA_PATTERN = re.compile(r'\d{2,3}[/-]\d{1,2}[/-]\d{1,2}')  # regex of republic era
    AD_PATTERN = re.compile(r'\d{4}[/-]\d{1,2}[/-]\d{1,2}')  # regex of ad


class StockPattern:
    CODE_NAME_PATTERN = re.compile(r'[\da-zA-Z]+\s.*')


class TradingPattern:
    DIFFERENT_PATTERN = re.compile(r'[+-]')


class MarginTradingPattern:
    NOTE_STRIP_PATTERN = re.compile(r'[\s]+')


class PERatioPattern:
    CHINESE_DATE_PATTERN = re.compile(r'([\d]+)年([\d]+)月([\d]+)日')



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from tw_stock_plugin.regex_pattern import MarginTradingPattern
from tw_stock_plugin.utils.schema import SchemaPattern


class TwseMarginTradingObject:
    def __init__(self, code, name, margin_purchase, margin_sells, cash_redemption, cash_balance_of_previous_day,
                 cash_balance_of_the_day, cash_quota, short_covering, short_sale, stock_redemption,
                 stock_balance_of_previous_day, stock_balance_of_the_day, stock_quota, offset, note):
        """
        :param code: 股票代碼
        :param name: 股票名稱
        :param margin_purchase: 融資買進
        :param margin_sells: 融資賣出
        :param cash_redemption: 現金償還
        :param cash_balance_of_previous_day: 前日餘額
        :param cash_balance_of_the_day: 今日餘額
        :param cash_quota: 限額
        :param short_covering: 融券買進
        :param short_sale: 融券賣出
        :param stock_redemption: 現金償還
        :param stock_balance_of_previous_day: 前日餘額
        :param stock_balance_of_the_day: 今日餘額
        :param stock_quota: 限額
        :param offset: 資券互抵
        :param note: 備註
        """
        self.code = code
        self.name = name
        self.margin_purchase = margin_purchase
        self.margin_sells = margin_sells
        self.cash_redemption = cash_redemption
        self.cash_balance_of_previous_day = cash_balance_of_previous_day
        self.cash_balance_of_the_day = cash_balance_of_the_day
        self.cash_quota = cash_quota
        self.short_covering = short_covering
        self.short_sale = short_sale
        self.stock_redemption = stock_redemption
        self.stock_balance_of_previous_day = stock_balance_of_previous_day
        self.stock_balance_of_the_day = stock_balance_of_the_day
        self.stock_quota = stock_quota
        self.offset = offset
        self.note = note
        self._format_value()
        # self._valid_schema()

    def _format_value(self):
        for key, value in self.__dict__.items():
            value = value.strip()
            if value and ',' in value:
                setattr(self, key, float(value.replace(',', '')))
            elif key in {'code', 'name'}:
                setattr(self, key, value)
            elif key in {'note'}:
                if MarginTradingPattern.NOTE_STRIP_PATTERN.search(value):
                    value = MarginTradingPattern.NOTE_STRIP_PATTERN.sub(',', value)
                    setattr(self, key, value)
                elif not getattr(self, key) or not value:
                    setattr(self, key, None)
                else:
                    setattr(self, key, value)
            else:
                setattr(self, key, float(value))

    def _valid_schema(self):
        SchemaPattern.TwseStockMarginTradingSchema.validate(self.__dict__)


class TpexMarginTradingObject:
    def __init__(self, code, name, cash_balance_of_previous_day, margin_purchase, margin_sells, cash_redemption,
                 cash_balance_of_the_day, cash_belong_to_securities_finance, cash_utilization_rate, cash_quota,
                 stock_balance_of_previous_day, short_covering, short_sale, stock_redemption, stock_balance_of_the_day,
                 stock_belong_to_securities_finance, stock_utilization_rate, stock_quota, offset, note):
        """
        :param code: 股票代碼
        :param name: 股票名稱
        :param cash_balance_of_previous_day: 前日餘額
        :param margin_purchase: 融資買進
        :param margin_sells: 融資賣出
        :param cash_redemption: 現金償還
        :param cash_balance_of_the_day: 今日餘額
        :param cash_belong_to_securities_finance: 資屬證金
        :param cash_utilization_rate(%): 資使用率(%)
        :param cash_quota: 限額
        :param stock_balance_of_previous_day: 前日餘額
        :param short_covering: 融券買進
        :param short_sale: 融券賣出
        :param stock_redemption: 現金償還
        :param stock_balance_of_the_day: 今日餘額
        :param stock_belong_to_securities_finance: 券屬證金
        :param stock_utilization_rate(%): 券使用率(%)
        :param stock_quota: 限額
        :param offset: 資券互抵
        :param note: 備註
        """
        self.code = code
        self.name = name
        self.cash_balance_of_previous_day = cash_balance_of_previous_day
        self.margin_purchase = margin_purchase
        self.margin_sells = margin_sells
        self.cash_redemption = cash_redemption
        self.cash_balance_of_the_day = cash_balance_of_the_day
        self.cash_belong_to_securities_finance = cash_belong_to_securities_finance
        self.cash_utilization_rate = cash_utilization_rate
        self.cash_quota = cash_quota
        self.stock_balance_of_previous_day = stock_balance_of_previous_day
        self.short_covering = short_covering
        self.short_sale = short_sale
        self.stock_redemption = stock_redemption
        self.stock_balance_of_the_day = stock_balance_of_the_day
        self.stock_belong_to_securities_finance = stock_belong_to_securities_finance
        self.stock_utilization_rate = stock_utilization_rate
        self.stock_quota = stock_quota
        self.offset = offset
        self.note = note
        self._format_value()
        # self._valid_schema()

    def _format_value(self):
        for key, value in self.__dict__.items():
            value = value.strip()
            if value and ',' in value:
                setattr(self, key, float(value.replace(',', '')))
            elif key in {'code', 'name'}:
                setattr(self, key, value)
            elif key in {'note'}:
                if MarginTradingPattern.NOTE_STRIP_PATTERN.search(value):
                    value = MarginTradingPattern.NOTE_STRIP_PATTERN.sub(',', value)
                    setattr(self, key, value)
                elif not getattr(self, key) or not value:
                    setattr(self, key, None)
                else:
                    setattr(self, key, value)
            else:
                setattr(self, key, float(value))

    def _valid_schema(self):
        SchemaPattern.TpexStockMarginTradingSchema.validate(self.__dict__)



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from datetime import date

from tw_stock_plugin.core.stock_tools import StockTools
from tw_stock_plugin.object.stock_margin_trading import TwseMarginTradingObject, TpexMarginTradingObject
from tw_stock_plugin.constant import Domain
from tw_stock_plugin.utils.response_handler import ResponseHandler


class StockMarginTrading:
    def __init__(self, date_):
        """
        :param date_: 查詢日期
        """
        self.date_ = date_
        self._valid_date()

    def _valid_date(self):
        """
        valid date format
        """
        if not isinstance(self.date_, date):
            raise TypeError('input date must be type of datetime.date')

    def _fetch_twse_data_all(self):
        """
        fetch all margin trading data in specific date from twse website
        :return:
        """
        margin_trading_dict = dict()
        url = f'{Domain.TAIWAN_STOCK_EXCHANGE_CORPORATION}/exchangeReport/MI_MARGN'
        query_date = self.date_.strftime('%Y%m%d')
        params = {
            'response': 'json',
            'date': query_date,
            'selectType': 'ALL'
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        stats = json_data['stat']
        if not stats == 'OK':
            return None
        data_list = json_data['data']
        columns = ['code', 'name', 'margin_purchase', 'margin_sells', 'cash_redemption', 'cash_balance_of_previous_day',
                   'cash_balance_of_the_day', 'cash_quota', 'short_covering', 'short_sale', 'stock_redemption',
                   'stock_balance_of_previous_day', 'stock_balance_of_the_day', 'stock_quota', 'offset', 'note']
        for data in data_list:
            stock_margin_trading = TwseMarginTradingObject(**dict(zip(columns, data)))
            margin_trading_dict[stock_margin_trading.code] = stock_margin_trading
        return margin_trading_dict

    def _fetch_tpex_data_all(self):
        """
        fetch all margin trading data in specific date from tpex website
        :return:
        """
        margin_trading_dict = dict()
        url = f'{Domain.TAIPEI_EXCHANGE}/web/stock/margin_trading/margin_balance/margin_bal_result.php'
        query_date = StockTools.ad_to_republic_era(date_=self.date_).replace('-', '/'),
        params = {
            'l': 'zh-tw',
            'o': 'json',
            'd': query_date
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        data_list = json_data['aaData']
        if not data_list:
            return None
        columns = ['code', 'name', 'cash_balance_of_previous_day', 'margin_purchase', 'margin_sells', 'cash_redemption',
                   'cash_balance_of_the_day', 'cash_belong_to_securities_finance', 'cash_utilization_rate',
                   'cash_quota', 'stock_balance_of_previous_day', 'short_covering', 'short_sale', 'stock_redemption',
                   'stock_balance_of_the_day', 'stock_belong_to_securities_finance', 'stock_utilization_rate',
                   'stock_quota', 'offset', 'note']
        for data in data_list:
            stock_margin_trading = TpexMarginTradingObject(**dict(zip(columns, data)))
            margin_trading_dict[stock_margin_trading.code] = stock_margin_trading
        return margin_trading_dict

    def get_all(self):
        """
        return all margin trading data
        :return:
        """
        margin_trading_dict = dict()
        twse_data = self._fetch_twse_data_all()
        tpex_data = self._fetch_tpex_data_all()
        if twse_data:
            margin_trading_dict.update(twse_data)
        if tpex_data:
            margin_trading_dict.update(tpex_data)
        return margin_trading_dict



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

from datetime import date
from schema import Schema, Or


class SchemaPattern:
    StockInfoSchema = Schema({
        'code': str,
        'name': str,
        'ISIN_code': str,
        'listed_date': str,
        'category': str,
        'industry': Or(str, None),
        'CFI_code': str,
        'type': str
    })

    TwseStockTradingSchema = Schema({
        'code': Or(str, None),
        'name': Or(str, None),
        'closing_price': Or(float, None),
        'change': Or(float, str, None),
        'opening_price': Or(float, None),
        'highest_price': Or(float, None),
        'lowest_price': Or(float, None),
        '_different': Or(str, None),
        'trading_volume': Or(float, None),
        'trade_value': Or(float, None),
        'transaction': Or(float, None),
        'last_best_bid_price': Or(float, None),
        'last_best_ask_price': Or(float, None),
        'last_best_bid_volume': Or(float, None),
        'last_best_ask_volume': Or(float, None),
        'price_earning_rate': Or(float, None),
    })

    TpexStockTradingSchema = Schema(
        {
            'code': Or(str, None),
            'name': Or(str, None),
            'closing_price': Or(float, None),
            'change': Or(float, str, None),
            'opening_price': Or(float, None),
            'highest_price': Or(float, None),
            'lowest_price': Or(float, None),
            'trading_volume': Or(float, None),
            'trade_value': Or(float, None),
            'transaction': Or(float, None),
            'last_best_bid_price': Or(float, None),
            'last_best_ask_price': Or(float, None),
            'last_best_bid_volume': Or(float, None),
            'last_best_ask_volume': Or(float, None),
            'issued_shares': Or(float, None),
            'next_limit_up': Or(float, None),
            'next_limit_down': Or(float, None),
        }
    )

    StockInstitutionalInvestorsSchema = Schema(
        {
            'code': str,
            'name': str,
            'foreign_mainland_area_buy': float,
            'foreign_mainland_area_sell': float,
            'foreign_mainland_area_diff': float,
            'foreign_buy': float,
            'foreign_sell': float,
            'foreign_diff': float,
            'trust_buy': float,
            'trust_sell': float,
            'trust_diff': float,
            'proprietary_dealers_buy': float,
            'proprietary_dealers_sell': float,
            'proprietary_dealers_diff': float,
            'hedge_dealers_buy': float,
            'hedge_dealers_sell': float,
            'hedge_dealers_diff': float,
            'total_diff': float,
        }
    )

    TwseStockMarginTradingSchema = Schema({
        'code': str,
        'name': str,
        'margin_purchase': float,
        'margin_sells': float,
        'cash_redemption': float,
        'cash_balance_of_previous_day': float,
        'cash_balance_of_the_day': float,
        'cash_quota': float,
        'short_covering': float,
        'short_sale': float,
        'stock_redemption': float,
        'stock_balance_of_previous_day': float,
        'stock_balance_of_the_day': float,
        'stock_quota': float,
        'offset': float,
        'note': Or(str, None)}
    )

    TpexStockMarginTradingSchema = Schema({
        'code': str,
        'name': str,
        'cash_balance_of_previous_day': float,
        'margin_purchase': float,
        'margin_sells': float,
        'cash_redemption': float,
        'cash_balance_of_the_day': float,
        'cash_belong_to_securities_finance': float,
        'cash_utilization_rate': float,
        'cash_quota': float,
        'stock_balance_of_previous_day': float,
        'short_covering': float,
        'short_sale': float,
        'stock_redemption': float,
        'stock_balance_of_the_day': float,
        'stock_belong_to_securities_finance': float,
        'stock_utilization_rate': float,
        'stock_quota': float,
        'offset': float,
        'note': Or(str, None)
    })

    StockShareholdingsSchema = Schema({
        'date': date,
        'code': str,
        'index': int,
        'number_of_shares': Or(str, None),
        'number_of_shareholders': int,
        'total_shares': int,
        'percentage_over_total_shares': float
    })

    StockPERatioSchema = Schema({
        'name': Or(str, None),
        'code': Or(str, None),
        'date': Or(date, None),
        'per': Or(float, None),
        'dividend_per_share': Or(float, None),
        'dividend_year': Or(int, None),
        'yield_ratio': Or(float, None),
        'pbr': Or(float, None),
        'fiscal_year_quarter': Or(str, None)
    })



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from datetime import datetime, date

from tw_stock_plugin.core.stock_tools import StockTools
from tw_stock_plugin.object.stock_trading import TwseTradingObject, TpexTradingObject
from tw_stock_plugin.constant import Domain
from tw_stock_plugin.utils.response_handler import ResponseHandler


class StockTrading:
    def __init__(self, date_):
        """
        :param date_: 查詢日期
        """
        self.date_ = date_
        self._valid_date()

    def _valid_date(self):
        """
        valid date format
        """
        if not isinstance(self.date_, date):
            raise TypeError('input date must be type of datetime.date')

    def _fetch_twse_data_all(self):
        """
        fetch all trading data in specific date from twse website
        :return:
        """
        trading_dict = dict()
        url = f'{Domain.TAIWAN_STOCK_EXCHANGE_CORPORATION}/exchangeReport/MI_INDEX'
        query_date = self.date_.strftime('%Y%m%d')
        params = {
            'response': 'json',
            'date': query_date,
            'type': 'ALL'
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        stats = json_data['stat']
        if not stats == 'OK':
            return None
        data_list = json_data['data9']
        columns = ['code', 'name', 'trading_volume', 'transaction', 'trade_value', 'opening_price', 'highest_price',
                   'lowest_price', 'closing_price', 'different', 'change', 'last_best_bid_price',
                   'last_best_bid_volume', 'last_best_ask_price', 'last_best_ask_volume', 'price_earning_rate']
        for data in data_list:
            stock_trading = TwseTradingObject(**dict(zip(columns, data)))
            trading_dict[stock_trading.code] = stock_trading
        return trading_dict

    def _fetch_tpex_data_all(self):
        """
        fetch all trading data in specific date from tpex website
        :return:
        """
        trading_dict = dict()
        url = f'{Domain.TAIPEI_EXCHANGE}/web/stock/aftertrading/otc_quotes_no1430/stk_wn1430_result.php'
        query_date = StockTools.ad_to_republic_era(date_=self.date_).replace('-', '/'),
        params = {
            'l': 'zh-tw',
            'd': query_date,
            'se': 'EW'
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        data_list = json_data['aaData']
        if not data_list:
            return None
        columns = ['code', 'name', 'closing_price', 'change', 'opening_price', 'highest_price', 'lowest_price',
                   'trading_volume', 'trade_value', 'transaction', 'last_best_bid_price', 'last_best_bid_volume',
                   'last_best_ask_price', 'last_best_ask_volume', 'issued_shares', 'next_limit_up', 'next_limit_down']
        for data in data_list:
            stock_trading = TpexTradingObject(**dict(zip(columns, data)))
            trading_dict[stock_trading.code] = stock_trading
        return trading_dict

    def get_all(self):
        """
        return all trading data
        :return:
        """
        trading_dict = dict()
        twse_data = self._fetch_twse_data_all()
        tpex_data = self._fetch_tpex_data_all()
        if twse_data:
            trading_dict.update(twse_data)
        if tpex_data:
            trading_dict.update(tpex_data)
        return trading_dict

    def _fetch_twse_data_history(self, code):
        """
        get specific stock monthly trading data from twse website
        :param code:
        :return:
        """
        trading_dict = dict()
        url = f'{Domain.TAIWAN_STOCK_EXCHANGE_CORPORATION}/exchangeReport/STOCK_DAY'
        query_date = self.date_.strftime('%Y%m%d')
        params = {
            'response': 'json',
            'date': query_date,
            'stockNo': code
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        stats = json_data['stat']
        if not stats == 'OK':
            return None
        data_list = json_data['data']
        columns = ['trading_volume', 'trade_value', 'opening_price', 'highest_price', 'lowest_price',
                   'closing_price', 'change', 'transaction']

        for data in data_list:
            date_ = StockTools.republic_era_to_ad(data[0])
            date_ = datetime.strptime(date_, '%Y/%m/%d').date()
            stock_trading = TwseTradingObject(**(dict(zip(columns, data[1:]))))
            for attr in stock_trading.__dict__.copy():
                if attr not in columns:
                    delattr(stock_trading, attr)
            trading_dict[date_] = stock_trading
        return trading_dict

    def _fetch_tpex_data_history(self, code):
        """
        get specific stock monthly trading data from tpex website
        :param code:
        :return:
        """
        trading_dict = dict()
        url = f'{Domain.TAIPEI_EXCHANGE}/web/stock/aftertrading/daily_trading_info/st43_result.php'
        query_date = StockTools.ad_to_republic_era(date_=self.date_).replace('-', '/'),
        params = {
            'l': 'zh-tw',
            'd': query_date,
            'stkno': code
        }
        response = ResponseHandler.get(url=url, params=params)
        if not response:
            return None
        json_data = response.json()
        data_list = json_data['aaData']
        if not data_list:
            return None
        columns = ['trading_volume', 'trade_value', 'opening_price', 'highest_price', 'lowest_price',
                   'closing_price', 'change', 'transaction']
        for data in data_list:
            date_ = StockTools.republic_era_to_ad(data[0])
            date_ = datetime.strptime(date_, '%Y/%m/%d').date()
            stock_trading = TpexTradingObject(**(dict(zip(columns, data[1:]))))
            for attr in stock_trading.__dict__.copy():
                if attr not in columns:
                    delattr(stock_trading, attr)
            trading_dict[date_] = stock_trading
        return trading_dict

    def get_history(self, code):
        """
        return monthly trading data with specific stock code
        :param code:
        :return:
        """
        twse_data = self._fetch_twse_data_history(code=code)
        if twse_data:
            return twse_data

        tpex_data = self._fetch_tpex_data_history(code=code)
        if tpex_data:
            return tpex_data

        raise ValueError(f'{code} {self.date_} result not found')



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from tw_stock_plugin.utils.schema import SchemaPattern


class StockPERatioObject:
    def __init__(self, yield_ratio, pbr, per, code=None, name=None, dividend_year=None, date_=None,
                 dividend_per_share=None,
                 fiscal_year_quarter=None):
        """
        :param code: 股票代碼
        :param name: 股票名稱
        :param date_: 資料日期
        :param per: 本益比
        :param dividend_per_share: 每股股利
        :param dividend_year: 股利年度
        :param yield_ratio: 殖利率(%)
        :param pbr: 股價淨值比
        :param fiscal_year_quarter: 財報年/季
        """
        self.code = code
        self.name = name
        self.date = date_
        self.per = per
        self.dividend_per_share = dividend_per_share
        self.dividend_year = dividend_year
        self.yield_ratio = yield_ratio
        self.pbr = pbr
        self.fiscal_year_quarter = fiscal_year_quarter
        self._format_value()
        # self._valid_schema()

    def _format_value(self):
        float_keys = {'per', 'dividend_per_share', 'yield_ratio', 'pbr'}
        for key in float_keys:
            value = getattr(self, key)
            if value in {'N/A', '-'}:
                setattr(self, key, None)
            elif isinstance(value, str) and ',' in value:
                setattr(self, key, value.replace(',', ''))

            value = getattr(self, key)
            if getattr(self, key):
                setattr(self, key, float(value))
        if self.dividend_year:
            self.dividend_year = int(self.dividend_year)

    def _valid_schema(self):
        SchemaPattern.StockPERatioSchema.validate(self.__dict__)



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from tw_stock_plugin.regex_pattern import TradingPattern
from tw_stock_plugin.utils.schema import SchemaPattern


class TwseTradingObject:
    def __init__(self, trading_volume, transaction, trade_value, opening_price, highest_price, lowest_price,
                 closing_price, change, code=None, name=None, different=None, last_best_bid_price=None,
                 last_best_bid_volume=None, last_best_ask_price=None, last_best_ask_volume=None,
                 price_earning_rate=None):

        """
        :param code: 股票代碼
        :param name: 股票名稱
        :param trading_volume: 成交股數
        :param transaction: 成交筆數
        :param trade_value: 成交金額
        :param opening_price: 開盤價
        :param highest_price: 最高價
        :param lowest_price: 最低價
        :param closing_price: 收盤價
        :param different: 漲跌(+/-)
        :param change: 漲跌價差
        :param last_best_bid_price: 最後揭示買價
        :param last_best_bid_volume: 最後揭示賣價
        :param last_best_ask_price: 最後揭示買量
        :param last_best_ask_volume: 最後揭示買量
        :param price_earning_rate: 本益比
        """
        self.code = code
        self.name = name
        self.trading_volume = trading_volume
        self.transaction = transaction
        self.trade_value = trade_value
        self.opening_price = opening_price
        self.highest_price = highest_price
        self.lowest_price = lowest_price
        self.closing_price = closing_price
        self._different = different
        self.change = change
        self.last_best_bid_price = last_best_bid_price
        self.last_best_bid_volume = last_best_bid_volume
        self.last_best_ask_price = last_best_ask_price
        self.last_best_ask_volume = last_best_ask_volume
        self.price_earning_rate = price_earning_rate
        self._format_value()
        # self._valid_schema()

    def _format_value(self):
        if self._different and TradingPattern.DIFFERENT_PATTERN.search(self._different):
            self._different = TradingPattern.DIFFERENT_PATTERN.search(self._different).group()
            self.change = f'{self._different}{self.change}' if self._different else self.change

        for key, value in self.__dict__.items():
            value = value.strip() if value else None
            if not value or key in {'name', 'code', '_different'}:
                continue
            elif value.startswith('X'):
                setattr(self, key, float(value.replace('X', '')))
            elif value.startswith('-') and value.endswith('-'):
                setattr(self, key, None)
            elif value and ',' in value:
                setattr(self, key, float(value.replace(',', '')))
            elif value in {'除息', '除權'}:
                setattr(self, key, value)
            else:
                setattr(self, key, float(value))

    def _valid_schema(self):
        SchemaPattern.TwseStockTradingSchema.validate(self.__dict__)


class TpexTradingObject:
    def __init__(self, closing_price, change, opening_price, highest_price, lowest_price, trading_volume, trade_value,
                 transaction, code=None, name=None, last_best_bid_price=None, last_best_ask_price=None,
                 last_best_bid_volume=None, last_best_ask_volume=None, issued_shares=None, next_limit_up=None,
                 next_limit_down=None):
        """
        :param code: 股票代碼
        :param name: 股票名稱
        :param closing_price: 收盤價
        :param change: 漲跌價差
        :param opening_price: 開盤價
        :param highest_price: 最高價
        :param lowest_price: 最低價
        :param trading_volume: 成交股數
        :param trade_value: 成交金額
        :param transaction: 成交筆數
        :param last_best_bid_price: 最後揭示買價
        :param last_best_ask_price: 最後揭示買量
        :param last_best_bid_volume: 最後揭示賣價
        :param last_best_ask_volume: 最後揭示買量
        :param issued_shares: 發行股數
        :param next_limit_up: 次日漲停價
        :param next_limit_down: 次日跌停價
        """
        self.code = code
        self.name = name
        self.closing_price = closing_price
        self.change = change
        self.opening_price = opening_price
        self.highest_price = highest_price
        self.lowest_price = lowest_price
        self.trading_volume = trading_volume
        self.trade_value = trade_value
        self.transaction = transaction
        self.last_best_bid_price = last_best_bid_price
        self.last_best_ask_price = last_best_ask_price
        self.last_best_bid_volume = last_best_bid_volume
        self.last_best_ask_volume = last_best_ask_volume
        self.issued_shares = issued_shares
        self.next_limit_up = next_limit_up
        self.next_limit_down = next_limit_down
        self._format_value()
        # self._valid_schema()

    def _format_value(self):
        for key, value in self.__dict__.items():
            value = value.strip() if value else None
            if not value or key in {'name', 'code', '_different'}:
                continue
            elif value.startswith('X'):
                setattr(self, key, float(value.replace('X', '')))
            elif (value.startswith('-') and value.endswith('-')) or not value:
                setattr(self, key, None)
            elif value and ',' in value:
                setattr(self, key, float(value.replace(',', '')))
            elif value in {'除息', '除權'}:
                setattr(self, key, value)
            else:
                setattr(self, key, float(value))

        if self.code and self.name:
            self.last_best_bid_volume = self.last_best_bid_volume * 1000 if self.last_best_bid_volume else None
            self.last_best_ask_volume = self.last_best_ask_volume * 1000 if self.last_best_ask_volume else None
        else:
            self.trading_volume = self.trading_volume * 1000 if self.trading_volume else None
            self.trade_value = self.trade_value * 1000 if self.trade_value else None

    def _valid_schema(self):
        SchemaPattern.TpexStockTradingSchema.validate(self.__dict__)



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
import unittest
import random
from datetime import datetime
from time import sleep

from tw_stock_plugin import StockInfo, StockTrading, StockInstitutionalInvestors, StockMarginTrading, \
    StockPERatio, StockShareholdings, StockTools, StockInfoObject


class TwStockPluginTest(unittest.TestCase):
    def setUp(self):
        self.testing_date = datetime(2020, 10, 30).date()
        self.stock_info = StockInfo()
        self.stock_trading = StockTrading(date_=self.testing_date)
        self.stock_institutional_investors = StockInstitutionalInvestors(date_=self.testing_date)
        self.stock_margin_trading = StockMarginTrading(date_=self.testing_date)
        self.stock_p_e_ratio = StockPERatio(date_=self.testing_date)
        self.stock_dict = {
            '2330': {'code': '2330',
                     'name': '台積電',
                     'ISIN_code': 'TW0002330008',
                     'listed_date': '1994/09/05',
                     'category': '上市',
                     'industry': '半導體業',
                     'CFI_code': 'ESVUFR',
                     'type': '股票'},
            '8446': {'code': '8446',
                     'name': '華研',
                     'ISIN_code': 'TW0008446006',
                     'listed_date': '2013/12/19',
                     'category': '上櫃',
                     'industry': '文化創意業',
                     'CFI_code': 'ESVUFR',
                     'type': '股票'},
            '1269': {'code': '1269',
                     'name': '乾杯',
                     'ISIN_code': 'TW0001269009',
                     'listed_date': '2016/08/29',
                     'category': '興櫃',
                     'industry': '觀光事業',
                     'CFI_code': 'ESVUFR',
                     'type': '股票'},
        }
        self.trading_history_dict = {
            '2330': {'trading_volume': 49770771,
                     'trade_value': 21614729093,
                     'opening_price': 437,
                     'highest_price': 437,
                     'lowest_price': 432,
                     'closing_price': 432,
                     'change': -5,
                     'transaction': 46871},
            '1101': {'trading_volume': 8723616,
                     'trade_value': 353231776,
                     'opening_price': 40.35,
                     'highest_price': 40.60,
                     'lowest_price': 40.30,
                     'closing_price': 40.55,
                     'change': 0.2,
                     'transaction': 2925},
            '9962': {'trading_volume': 59000,
                     'trade_value': 536000,
                     'opening_price': 9.02,
                     'highest_price': 9.11,
                     'lowest_price': 8.91,
                     'closing_price': 9.11,
                     'change': 0.09,
                     'transaction': 22}
        }
        self.trading_all_dict = {
            '1101': {'code': '1101',
                     'name': '台泥',
                     'trading_volume': 8723616,
                     'transaction': 2925,
                     'trade_value': 353231776,
                     'opening_price': 40.35,
                     'highest_price': 40.6,
                     'lowest_price': 40.3,
                     'closing_price': 40.55,
                     'change': 0.2,
                     'last_best_bid_price': 40.5,
                     'last_best_bid_volume': 59,
                     'last_best_ask_price': 40.55,
                     'last_best_ask_volume': 145,
                     'price_earning_rate': 9.56},
            '2330': {'code': '2330',
                     'name': '台積電',
                     'trading_volume': 49770771,
                     'transaction': 46871,
                     'trade_value': 21614729093,
                     'opening_price': 437,
                     'highest_price': 437,
                     'lowest_price': 432,
                     'closing_price': 432,
                     'change': -5,
                     'last_best_bid_price': 432,
                     'last_best_bid_volume': 1338,
                     'last_best_ask_price': 432.50,
                     'last_best_ask_volume': 3,
                     'price_earning_rate': 24.63},
            '1258': {'code': '1258',
                     'name': '其祥-KY',
                     'closing_price': 12.80,
                     'change': -0.3,
                     'opening_price': 12.8,
                     'highest_price': 12.8,
                     'lowest_price': 12.8,
                     'trading_volume': 2000,
                     'trade_value': 25600,
                     'transaction': 2,
                     'last_best_bid_price': 12.8,
                     'last_best_bid_volume': 3000,
                     'last_best_ask_price': 13.1,
                     'last_best_ask_volume': 1000,
                     'issued_shares': 36819791,
                     'next_limit_up': 14.05,
                     'next_limit_down': 11.55}
        }
        self.institutional_investors_all_dict = {
            '1101': {'code': '1101',
                     'name': '台泥',
                     'foreign_mainland_area_buy': 5474400,
                     'foreign_mainland_area_sell': 4865925,
                     'foreign_mainland_area_diff': 608475,
                     'foreign_buy': 0,
                     'foreign_sell': 0,
                     'foreign_diff': 0,
                     'trust_buy': 0,
                     'trust_sell': 0,
                     'trust_diff': 0,
                     'proprietary_dealers_buy': 14000,
                     'proprietary_dealers_sell': 516000,
                     'proprietary_dealers_diff': -502000,
                     'hedge_dealers_buy': 835000,
                     'hedge_dealers_sell': 79000,
                     'hedge_dealers_diff': 756000,
                     'total_diff': 862475},
            '2330': {'code': '2330',
                     'name': '台積電',
                     'foreign_mainland_area_buy': 19943461,
                     'foreign_mainland_area_sell': 39095735,
                     'foreign_mainland_area_diff': -19152274,
                     'foreign_buy': 0,
                     'foreign_sell': 0,
                     'foreign_diff': 0,
                     'trust_buy': 30000,
                     'trust_sell': 46000,
                     'trust_diff': -16000,
                     'proprietary_dealers_buy': 75000,
                     'proprietary_dealers_sell': 1197000,
                     'proprietary_dealers_diff': -1122000,
                     'hedge_dealers_buy': 174000,
                     'hedge_dealers_sell': 431998,
                     'hedge_dealers_diff': -257998,
                     'total_diff': -20548272},
            '3260': {'code': '3260',
                     'name': '威剛',
                     'foreign_mainland_area_buy': 105000,
                     'foreign_mainland_area_sell': 285000,
                     'foreign_mainland_area_diff': -180000,
                     'foreign_buy': 0,
                     'foreign_sell': 0,
                     'foreign_diff': 0,
                     'trust_buy': 0,
                     'trust_sell': 0,
                     'trust_diff': 0,
                     'proprietary_dealers_buy': 14000,
                     'proprietary_dealers_sell': 30000,
                     'proprietary_dealers_diff': -16000,
                     'hedge_dealers_buy': 2000,
                     'hedge_dealers_sell': 18000,
                     'hedge_dealers_diff': -16000,
                     'total_diff': -212000}
        }
        self.margin_trading_all_dict = {
            # '1101': {
            #     'code': '1101',
            #     'name': '台泥',
            #     'margin_purchase': 42,
            #     'margin_sells': 93,
            #     'cash_redemption': 0,
            #     'cash_balance_of_previous_day': 9568,
            #     'cash_balance_of_the_day': 9517,
            #     'cash_quota': 1434625,
            #     'short_covering': 0,
            #     'short_sale': 17,
            #     'stock_redemption': 0,
            #     'stock_balance_of_previous_day': 178,
            #     'stock_balance_of_the_day': 195,
            #     'stock_quota': 1434625,
            #     'offset': 0,
            #     'note': None
            # },
            '00672L': {
                'code': '00672L',
                'name': '元大S&P原油正2',
                'margin_purchase': 0,
                'margin_sells': 0,
                'cash_redemption': 235,
                'cash_balance_of_previous_day': 950,
                'cash_balance_of_the_day': 715,
                'cash_quota': 0,
                'short_covering': 0,
                'short_sale': 0,
                'stock_redemption': 1,
                'stock_balance_of_previous_day': 1,
                'stock_balance_of_the_day': 0,
                'stock_quota': 0,
                'offset': 0,
                'note': 'OX'
            },
            # '1595': {
            #     'code': '1595',
            #     'name': '川寶',
            #     'cash_balance_of_previous_day': 194,
            #     'margin_purchase': 0,
            #     'margin_sells': 3,
            #     'cash_redemption': 0,
            #     'cash_balance_of_the_day': 191,
            #     'cash_belong_to_securities_finance': 6,
            #     'cash_utilization_rate': 1.62,
            #     'cash_quota': 11787,
            #     'stock_balance_of_previous_day': 0,
            #     'short_covering': 0,
            #     'short_sale': 0,
            #     'stock_redemption': 0,
            #     'stock_balance_of_the_day': 0,
            #     'stock_belong_to_securities_finance': 0,
            #     'stock_utilization_rate': 0,
            #     'stock_quota': 11787,
            #     'offset': 0,
            #     'note': '11,C'
            # }
        }
        self.p_e_ratio_history_dict = {
            '2330': {'yield_ratio': 2.2,
                     'dividend_year': 108,
                     'per': 24.63,
                     'pbr': 6.51,
                     'fiscal_year_quarter': '109/2'},
            '1101': {'yield_ratio': 7.40,
                     'dividend_year': 108,
                     'per': 9.56,
                     'pbr': 1.25,
                     'fiscal_year_quarter': '109/2'},
            '9962': {'per': None,
                     'yield_ratio': 2.2,
                     'dividend_year': 108,
                     'pbr': 0.89}
        }
        self.p_e_ratio_all_dict = {
            '1101': {'code': '1101',
                     'name': '台泥',
                     'yield_ratio': 7.4,
                     'dividend_year': 108,
                     'per': 9.56,
                     'pbr': 1.25,
                     'fiscal_year_quarter': '109/2'},
            '2330': {'code': '2330',
                     'name': '台積電',
                     'yield_ratio': 2.2,
                     'dividend_year': 108,
                     'per': 24.63,
                     'pbr': 6.51,
                     'fiscal_year_quarter': '109/2'},
            '2221': {'code': '2221',
                     'name': '大甲',
                     'per': 13.82,
                     'dividend_per_share': 1.5,
                     'dividend_year': 108,
                     'yield_ratio': 6.38,
                     'pbr': 1.33}
        }
        self.shareholdings_dict = {
            1: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 1,
                'number_of_shareholders': 78300,
                'number_of_shares': '1-999',
                'percentage_over_total_shares': 1.92,
                'total_shares': 17839119},
            2: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 2,
                'number_of_shareholders': 77229,
                'number_of_shares': '1000-5000',
                'percentage_over_total_shares': 15.55,
                'total_shares': 143973016},
            3: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 3,
                'number_of_shareholders': 7311,
                'number_of_shares': '5001-10000',
                'percentage_over_total_shares': 6.06,
                'total_shares': 56155961},
            4: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 4,
                'number_of_shareholders': 1974,
                'number_of_shares': '10001-15000',
                'percentage_over_total_shares': 2.72,
                'total_shares': 25262584},
            5: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 5,
                'number_of_shareholders': 1083,
                'number_of_shares': '15001-20000',
                'percentage_over_total_shares': 2.14,
                'total_shares': 19811145},
            6: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 6,
                'number_of_shareholders': 866,
                'number_of_shares': '20001-30000',
                'percentage_over_total_shares': 2.36,
                'total_shares': 21886879},
            7: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 7,
                'number_of_shareholders': 378,
                'number_of_shares': '30001-40000',
                'percentage_over_total_shares': 1.44,
                'total_shares': 13356254},
            8: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 8,
                'number_of_shareholders': 211,
                'number_of_shares': '40001-50000',
                'percentage_over_total_shares': 1.05,
                'total_shares': 9763707},
            9: {'code': '0050',
                'date': datetime(2020, 11, 6).date(),
                'index': 9,
                'number_of_shareholders': 372,
                'number_of_shares': '50001-100000',
                'percentage_over_total_shares': 2.85,
                'total_shares': 26397673},
            10: {'code': '0050',
                 'date': datetime(2020, 11, 6).date(),
                 'index': 10,
                 'number_of_shareholders': 132,
                 'number_of_shares': '100001-200000',
                 'percentage_over_total_shares': 1.96,
                 'total_shares': 18216225},
            11: {'code': '0050',
                 'date': datetime(2020, 11, 6).date(),
                 'index': 11,
                 'number_of_shareholders': 69,
                 'number_of_shares': '200001-400000',
                 'percentage_over_total_shares': 2.17,
                 'total_shares': 20105691},
            12: {'code': '0050',
                 'date': datetime(2020, 11, 6).date(),
                 'index': 12,
                 'number_of_shareholders': 14,
                 'number_of_shares': '400001-600000',
                 'percentage_over_total_shares': 0.74,
                 'total_shares': 6887616},
            13: {'code': '0050',
                 'date': datetime(2020, 11, 6).date(),
                 'index': 13,
                 'number_of_shareholders': 9,
                 'number_of_shares': '600001-800000',
                 'percentage_over_total_shares': 0.69,
                 'total_shares': 6452000},
            14: {'code': '0050',
                 'date': datetime(2020, 11, 6).date(),
                 'index': 14,
                 'number_of_shareholders': 9,
                 'number_of_shares': '800001-1000000',
                 'percentage_over_total_shares': 0.91,
                 'total_shares': 8434000},
            15: {'code': '0050',
                 'date': datetime(2020, 11, 6).date(),
                 'index': 15,
                 'number_of_shareholders': 55,
                 'number_of_shares': '1,000,001以上',
                 'percentage_over_total_shares': 57.36,
                 'total_shares': 530958130}
        }
        self.shareholdings_dict_all = {
            1: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 1,
                'number_of_shareholders': 80620,
                'number_of_shares': '1-999',
                'percentage_over_total_shares': 1.95,
                'total_shares': 18159107},
            2: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 2,
                'number_of_shareholders': 74015,
                'number_of_shares': '1,000-5,000',
                'percentage_over_total_shares': 14.81,
                'total_shares': 137930474},
            3: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 3,
                'number_of_shareholders': 6953,
                'number_of_shares': '5,001-10,000',
                'percentage_over_total_shares': 5.73,
                'total_shares': 53365960},
            4: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 4,
                'number_of_shareholders': 1902,
                'number_of_shares': '10,001-15,000',
                'percentage_over_total_shares': 2.61,
                'total_shares': 24308986},
            5: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 5,
                'number_of_shareholders': 1021,
                'number_of_shares': '15,001-20,000',
                'percentage_over_total_shares': 2.0,
                'total_shares': 18634713},
            6: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 6,
                'number_of_shareholders': 822,
                'number_of_shares': '20,001-30,000',
                'percentage_over_total_shares': 2.22,
                'total_shares': 20754627},
            7: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 7,
                'number_of_shareholders': 344,
                'number_of_shares': '30,001-40,000',
                'percentage_over_total_shares': 1.3,
                'total_shares': 12161242},
            8: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 8,
                'number_of_shareholders': 210,
                'number_of_shares': '40,001-50,000',
                'percentage_over_total_shares': 1.04,
                'total_shares': 9721246},
            9: {'code': '0050',
                'date': datetime(2020, 11, 13).date(),
                'index': 9,
                'number_of_shareholders': 355,
                'number_of_shares': '50,001-100,000',
                'percentage_over_total_shares': 2.7,
                'total_shares': 25229118},
            10: {'code': '0050',
                 'date': datetime(2020, 11, 13).date(),
                 'index': 10,
                 'number_of_shareholders': 128,
                 'number_of_shares': '100,001-200,000',
                 'percentage_over_total_shares': 1.89,
                 'total_shares': 17683736},
            11: {'code': '0050',
                 'date': datetime(2020, 11, 13).date(),
                 'index': 11,
                 'number_of_shareholders': 68,
                 'number_of_shares': '200,001-400,000',
                 'percentage_over_total_shares': 2.12,
                 'total_shares': 19824803},
            12: {'code': '0050',
                 'date': datetime(2020, 11, 13).date(),
                 'index': 12,
                 'number_of_shareholders': 16,
                 'number_of_shares': '400,001-600,000',
                 'percentage_over_total_shares': 0.82,
                 'total_shares': 7722616},
            13: {'code': '0050',
                 'date': datetime(2020, 11, 13).date(),
                 'index': 13,
                 'number_of_shareholders': 6,
                 'number_of_shares': '600,001-800,000',
                 'percentage_over_total_shares': 0.46,
                 'total_shares': 4310000},
            14: {'code': '0050',
                 'date': datetime(2020, 11, 13).date(),
                 'index': 14,
                 'number_of_shareholders': 10,
                 'number_of_shares': '800,001-1,000,000',
                 'percentage_over_total_shares': 0.96,
                 'total_shares': 9001000},
            15: {'code': '0050',
                 'date': datetime(2020, 11, 13).date(),
                 'index': 15,
                 'number_of_shareholders': 56,
                 'number_of_shares': '1,000,001以上',
                 'percentage_over_total_shares': 59.31,
                 'total_shares': 552192372}}
        self.date_convert_dict = {
            '2020/10/10': '109/10/10',
            '2020/01/01': '109/01/01',
            '1996/07/19': '85/07/19',
            '2020-10-10': '109-10-10',
            '2020-01-01': '109-01-01',
            '1996-07-19': '85-07-19'
        }

    def test_get_single(self):
        stock_code = random.choice(list(self.stock_dict))
        stock_info = StockInfoObject(**self.stock_dict[stock_code])
        stock_info_test = self.stock_info.get(code=stock_code)
        self.assertEqual(stock_info.__dict__, stock_info_test.__dict__)
        sleep(3)

    def test_get_stock_attribute(self):
        stock_code = random.choice(list(self.stock_dict))
        stock_info = self.stock_dict[stock_code]
        stock_info_test = self.stock_info.get(code=stock_code)
        for key in stock_info.keys():
            value = stock_info[key]
            value_test = getattr(stock_info_test, key)
            self.assertEqual(value, value_test)
        sleep(3)

    def test_is_open_date(self):
        open_day_date = datetime.strptime('2020/10/23', '%Y/%m/%d').date()
        holiday_date = datetime.strptime('2020/10/10', '%Y/%m/%d').date()
        self.assertTrue(StockTools.check_is_open_date(open_day_date))
        self.assertFalse(StockTools.check_is_open_date(holiday_date))
        sleep(3)

    def test_republic_era_to_ad(self):
        ad = random.choice(list(self.date_convert_dict))
        republic_era = self.date_convert_dict[ad]
        self.assertEqual(ad, StockTools.republic_era_to_ad(date_=republic_era))
        sleep(3)

    def test_ad_to_republic_era(self):
        ad = random.choice(list(self.date_convert_dict))
        republic_era = self.date_convert_dict[ad]
        self.assertEqual(republic_era, StockTools.ad_to_republic_era(date_=ad))
        sleep(3)

    def test_trading_data_history(self):
        stock_code = random.choice(list(self.trading_history_dict))
        stock_trading_history = self.trading_history_dict[stock_code]
        stock_trading_history_test = self.stock_trading.get_history(code=stock_code)[self.testing_date]
        for key in stock_trading_history.keys():
            value = stock_trading_history[key]
            value_test = getattr(stock_trading_history_test, key)
            self.assertEqual(value, value_test)
        sleep(3)

    def test_trading_data_all(self):
        stock_code = random.choice(list(self.trading_all_dict))
        stock_trading_all = self.trading_all_dict[stock_code]
        stock_trading_all_test = self.stock_trading.get_all()[stock_code]
        for key in stock_trading_all.keys():
            value = stock_trading_all[key]
            value_test = getattr(stock_trading_all_test, key)
            self.assertEqual(value, value_test)
        sleep(3)

    def test_institutional_investors_data_all(self):
        stock_code = random.choice(list(self.institutional_investors_all_dict))
        stock_institutional_investors_all = self.institutional_investors_all_dict[stock_code]
        stock_institutional_investors_all_test = self.stock_institutional_investors.get_all()[stock_code]
        for key in stock_institutional_investors_all.keys():
            value = stock_institutional_investors_all[key]
            value_test = getattr(stock_institutional_investors_all_test, key)
            self.assertEqual(value, value_test)
        sleep(3)

    def test_margin_trading_data_all(self):
        stock_code = random.choice(list(self.margin_trading_all_dict))
        stock_margin_trading_all = self.margin_trading_all_dict[stock_code]
        stock_margin_trading_all_test = self.stock_margin_trading.get_all()[stock_code]
        for key in stock_margin_trading_all.keys():
            value = stock_margin_trading_all[key]
            value_test = getattr(stock_margin_trading_all_test, key)
            self.assertEqual(value, value_test)
        sleep(3)

    def test_p_e_ratio_data_history(self):
        stock_code = random.choice(list(self.p_e_ratio_history_dict))
        stock_p_e_ratio = self.p_e_ratio_history_dict[stock_code]
        stock_p_e_ratio_test = self.stock_p_e_ratio.get_history(code=stock_code)[self.testing_date]
        for key in stock_p_e_ratio.keys():
            value = stock_p_e_ratio[key]
            value_test = getattr(stock_p_e_ratio_test, key)
            self.assertEqual(value, value_test)
        sleep(3)

    def test_p_e_ratio_data_all(self):
        stock_code = random.choice(list(self.p_e_ratio_all_dict))
        stock_p_e_ratio_all = self.p_e_ratio_all_dict[stock_code]
        stock_p_e_ratio_all_test = self.stock_p_e_ratio.get_all()[stock_code]
        for key in stock_p_e_ratio_all.keys():
            value = stock_p_e_ratio_all[key]
            value_test = getattr(stock_p_e_ratio_all_test, key)
            self.assertEqual(value, value_test)
        sleep(3)

    def test_shareholdings_data(self):
        stock_code = '0050'
        date_ = datetime(2020, 11, 6).date()
        stock_shareholdings = StockShareholdings()
        stock_shareholdings_test = stock_shareholdings.get_by_query(code=stock_code,
                                                                    date_=date_)
        for index in self.shareholdings_dict.keys():
            dict_ = self.shareholdings_dict[index]
            dict_test = stock_shareholdings_test[index]
            for key in dict_.keys():
                value = dict_[key]
                value_test = getattr(dict_test, key)
                self.assertEqual(value, value_test)
        sleep(3)

    def test_shareholdings_data_all(self):
        stock_code = '0050'
        stock_shareholdings = StockShareholdings()
        stock_shareholdings_test = stock_shareholdings.get_newest()[stock_code]
        for index in self.shareholdings_dict_all.keys():
            dict_ = self.shareholdings_dict_all[index]
            dict_test = stock_shareholdings_test[index]
            for key in dict_.keys():
                value = dict_[key]
                value_test = getattr(dict_test, key)
                self.assertEqual(value, value_test)
        sleep(3)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

import requests


class ResponseHandler:
    @classmethod
    def get(cls, url, params=None, headers=None, timeout=10):
        response = requests.get(url, params=params, headers=headers, timeout=timeout)
        return response

    @classmethod
    def post(cls, url, params=None, payload=None, headers=None, timeout=10):
        response = requests.post(url=url, params=params, data=payload, headers=headers, timeout=timeout)
        return response



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

from tw_stock_plugin.core.update_stock import *
from tw_stock_plugin.core.stock_tools import *
from tw_stock_plugin.core.stock_info import *
from tw_stock_plugin.core.stock_trading import *
from tw_stock_plugin.core.stock_institutional_investors import *
from tw_stock_plugin.core.stock_margin_trading import *
from tw_stock_plugin.core.stock_peratio import *
from tw_stock_plugin.core.stock_shareholdings import *



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

from datetime import datetime, date

from tw_stock_plugin.regex_pattern import DatePattern
from tw_stock_plugin.constant import Domain
from tw_stock_plugin.utils.response_handler import ResponseHandler


class StockTools:

    @classmethod
    def check_is_open_date(cls, date_):
        """
        check if input date is open date (available in after-hour time)
        :param date_:
        :return:
        """
        if not isinstance(date_, date):
            raise TypeError('input date must be type of datetime.date')
        if date_ > datetime.now().date():
            raise AttributeError('The input date must earlier than today.'.format(date))

        query_date = cls.ad_to_republic_era(date_=date_).replace('-', '/')
        url = f'{Domain.TAIPEI_EXCHANGE}/web/stock/aftertrading/index_summary/summary_result.php'
        params = {'l': 'zh-tw',
                  'd': query_date}
        response = ResponseHandler.get(url=url, params=params)
        json_data = response.json()['aaData']
        if json_data:
            return True
        else:
            return False

    @classmethod
    def ad_to_republic_era(cls, date_):
        """
        convert ad to republic era
            - 2020/10/10 -> 109/10/10
        :param date_:
        :return:
        """
        if isinstance(date_, date):
            date_ = date_.strftime('%Y-%m-%d')

        if DatePattern.AD_PATTERN.fullmatch(date_) and '/' in date_:
            year, month, day = date_.split('/')
            return '{}/{:02d}/{:02d}'.format(int(year) - 1911,
                                             int(month),
                                             int(day))
        elif DatePattern.AD_PATTERN.fullmatch(date_) and '-' in date_:
            year, month, day = date_.split('-')
            return '{}-{:02d}-{:02d}'.format(int(year) - 1911,
                                             int(month),
                                             int(day))
        else:
            raise ValueError('date format error, it is not ad format')

    @classmethod
    def republic_era_to_ad(cls, date_):
        """
        convert republic era to ad
            - 109/10/10 -> 2020/10/10
        :param date_:
        :return:
        """
        if DatePattern.REPUBLIC_ERA_PATTERN.fullmatch(date_) and '/' in date_:
            year, month, day = date_.split('/')
            return '{}/{:02d}/{:02d}'.format(int(year) + 1911,
                                             int(month),
                                             int(day))
        elif DatePattern.REPUBLIC_ERA_PATTERN.fullmatch(date_) and '-' in date_:
            year, month, day = date_.split('-')
            return '{}-{:02d}-{:02d}'.format(int(year) + 1911,
                                             int(month),
                                             int(day))
        else:
            raise ValueError('date format error, it is not republic era format')



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
from tw_stock_plugin.utils.schema import SchemaPattern


class InstitutionalInvestorsObject:
    def __init__(self, code, name, foreign_mainland_area_buy, foreign_mainland_area_sell, foreign_mainland_area_diff,
                 foreign_buy, foreign_sell, foreign_diff, trust_buy, trust_sell, trust_diff, proprietary_dealers_buy,
                 proprietary_dealers_sell, proprietary_dealers_diff, hedge_dealers_buy, hedge_dealers_sell,
                 hedge_dealers_diff, total_diff):
        """
        :param code: 股票代碼
        :param name: 股票名稱
        :param foreign_mainland_area_buy: 外陸資買進股數(不含外資自營商)
        :param foreign_mainland_area_sell: 外陸資賣出股數(不含外資自營商)
        :param foreign_mainland_area_diff: 外陸資買賣超股數(不含外資自營商)
        :param foreign_buy: 外資自營商買進股數
        :param foreign_sell: 外資自營商賣出股數
        :param foreign_diff: 外資自營商買賣超股數
        :param trust_buy: 投信買進股數
        :param trust_sell: 投信賣出股數
        :param trust_diff: 投信買賣超股數
        :param proprietary_dealers_buy: 自營商買進股數(自行買賣)
        :param proprietary_dealers_sell: 自營商賣出股數(自行買賣)
        :param proprietary_dealers_diff: 自營商買賣超股數(自行買賣)
        :param hedge_dealers_buy: 自營商買進股數(避險)
        :param hedge_dealers_sell: 自營商賣出股數(避險)
        :param hedge_dealers_diff: 自營商買賣超股數(避險)
        :param total_diff: 三大法人買賣超股數
        """
        self.code = code
        self.name = name
        self.foreign_mainland_area_buy = foreign_mainland_area_buy
        self.foreign_mainland_area_sell = foreign_mainland_area_sell
        self.foreign_mainland_area_diff = foreign_mainland_area_diff
        self.foreign_buy = foreign_buy
        self.foreign_sell = foreign_sell
        self.foreign_diff = foreign_diff
        self.trust_buy = trust_buy
        self.trust_sell = trust_sell
        self.trust_diff = trust_diff
        self.proprietary_dealers_buy = proprietary_dealers_buy
        self.proprietary_dealers_sell = proprietary_dealers_sell
        self.proprietary_dealers_diff = proprietary_dealers_diff
        self.hedge_dealers_buy = hedge_dealers_buy
        self.hedge_dealers_sell = hedge_dealers_sell
        self.hedge_dealers_diff = hedge_dealers_diff
        self.total_diff = total_diff
        self._format_value()
        # self._valid_schema()

    def _format_value(self):
        for key, value in self.__dict__.items():
            value = value.strip()
            if value and ',' in value:
                setattr(self, key, float(value.replace(',', '')))
            elif key in {'code', 'name'}:
                setattr(self, key, value)
            else:
                setattr(self, key, float(value))

    def _valid_schema(self):
        SchemaPattern.StockInstitutionalInvestorsSchema.validate(self.__dict__)



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""
import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='tw-stock-plugin',
    version='0.0.9',
    author='Jimmy Yeh',
    author_email='chienfeng0719@hotmail.com',
    description='Some util function when doing Taiwan stock web scraping and some common stock data parser',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/chienfeng0719/tw-stock-plugin',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'pandas==1.1.0',
        'numpy==1.19.1',
        'requests==2.24.0',
        'lxml==4.5.2',
        'schema==0.7.2'
    ],
    entry_points={
        'console_scripts': [
        ]
    },
    python_requires='>=3.6',
)



# -*- coding: utf-8 -*
"""
      ┏┓       ┏┓
    ┏━┛┻━━━━━━━┛┻━┓
    ┃      ☃      ┃
    ┃  ┳┛     ┗┳  ┃
    ┃      ┻      ┃
    ┗━┓         ┏━┛
      ┗┳        ┗━┓
       ┃          ┣┓
       ┃          ┏┛
       ┗┓┓┏━━━━┳┓┏┛
        ┃┫┫    ┃┫┫
        ┗┻┛    ┗┻┛
    God Bless,Never Bug
"""

from datetime import datetime

from tw_stock_plugin.constant import StockShareholdings
from tw_stock_plugin.utils.schema import SchemaPattern


class StockShareholdingsObject:
    def __init__(self, date, code, index, number_of_shareholders, total_shares,
                 percentage_over_total_shares, number_of_shares=None):
        """
        :param date: 日期
        :param code: 股票代碼
        :param index: 序
        :param number_of_shares: 持股/單位數分級
        :param number_of_shareholders: 人數
        :param total_shares: 股數/單位數
        :param percentage_over_total_shares: 占集保庫存數比例 (%)
        :return:
        """
        self.date = datetime.strptime(str(date), '%Y%m%d').date()
        self.code = code
        self.index = index
        self.number_of_shares = number_of_shares or self._transfer_number_of_shares()
        self.number_of_shareholders = number_of_shareholders
        self.total_shares = total_shares
        self.percentage_over_total_shares = percentage_over_total_shares
        self._format_value()
        # self._valid_schema()

    def _transfer_number_of_shares(self):
        if self.index not in StockShareholdings.NUMBER_OF_SHARES_DICT:
            return None
        return StockShareholdings.NUMBER_OF_SHARES_DICT[self.index]

    def _format_value(self):
        integer_keys = {'index', 'number_of_shareholders', 'total_shares'}
        float_keys = {'percentage_over_total_shares'}
        for key in integer_keys:
            value = getattr(self, key)
            setattr(self, key, int(value))

        for key in float_keys:
            value = getattr(self, key)
            setattr(self, key, float(value))

    def _valid_schema(self):
        SchemaPattern.StockShareholdingsSchema.validate(self.__dict__)
